/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>
#include "tcpjunk.h"
#include "gui.h"
#include "utils.h"

void tj_show_help()
  {
  printf("Tcpjunk v%s - TCP protocols testing tool\n", VER);
  puts(
	  "\nGeneral options:\n\
  -d, --debug              force display of session data when using instances\n\
                           or in server mode\n\
  -X, --hexdump            show session data as hex dump\n\
  -v, --verbose            display warnings and other information\n\
  -h, --help               show this help and exit\n\
  -x, --gui                run Tcpjunk in GUI mode\n\
  -a, --daemon             run Tcpjunk as a daemon\n\n\
Session options:\n\
  -c, --file=FILENAME        session data file\n\
  -e, --seq-file=DIRECTORY   use session files from directory, sequentially\n\
  -f, --rand-file=DIRECTORY  use random session files from directory\n\
  -W, --wait-data            wait for data at first\n\
  -C, --close                do not wait for data at last\n\
  -T, --sess-time=MSECONDS   session timeout (default: 100 milliseconds)\n\n\
Client options:\n\
  -s, --host=HOST/IP         hostname/IP of remote host\n\
  -i, --instances=NUMBER     number of client instances to run in parallel\n\
  -l, --limit-sess=NUMBER    limit number of sessions per instance (default:\n\
                             no limit)\n\
  -P, --src-ip=IP            bind to specific/random source IP addresses\n\
  -I, --interface=INTERFACE  network interface for source IP usage\n\
  -N, --ip-sessions=NUMBER   number of sessions to use with each source IP\n\
                             (default: 10)\n\
  -O, --offset=NUMBER        source IPv4 interface alias offset (default: 0)\n\
  -o, --src-port=PORT        bind to specific source port\n\n\
Server options:\n\
  -r, --server            server mode\n\
  -E, --cert=FILENAME     server certificate file\n\
  -K, --key=FILENAME      server key file\n\n\
Connection options:\n\
  -p, --port=PORT            service name/port to connect or listen to\n\
  -6, --ipv6                 Use IPv6\n\
  -S, --ssl                  use SSL v3.0\n\
  -t, --timeout=SECONDS      connection timeout (default: 2 seconds)\n\
  -w, --wait=MSECONDS        delay between connections (default: none)\n\
  -k, --repeat-sess=NUMBER   repeat session on same connection (default: 1)\n\
  -m, --send-size=BYTES      set send buffer size\n\
  -D, --send-delay=MSECONDS  delay before send operations\n\
  -z, --optimize             optimize TCP tw_recycle, tw_reuse settings and\n\
                             file descriptor limits\n\n\
See manual page for session file syntax and other information.\n\
Bug reports and suggestions: <oferkv@gmail.com>");

  exit(EXIT_FAILURE);
  }

void sigchld_handler(int s)
  {
  while (wait(NULL) > 0)
	;
  }

long get_file_size(FILE *stream)
  {
  long curpos, len;
  curpos = ftell(stream);
  fseek(stream, 0L, SEEK_END);
  len = ftell(stream);
  fseek(stream, curpos, SEEK_SET);
  return len;
  }

int load_session_data(struct tj_data *inst_data)
  {
  /* load session file */
  FILE *sess_fp;
  if ((sess_fp = fopen(inst_data->sess_file_name, "r")) == NULL)
	{
	sprintf(log_msg, "Failed to open session file %s: %s", inst_data->sess_file_name,
		strerror(errno));
	tj_log(inst_data, log_msg, c_red);
	if (inst_data->gui == 0)
	  exit(EXIT_FAILURE);
	else
	  return -1;
	}

  /* set session length */
  inst_data->sess_data_len = get_file_size(sess_fp);
  if (inst_data->sess_data_len < 1)
	{
	if (inst_data->gui == 0)
	  {
	  sprintf(log_msg, "the file %s is empty, terminating...",
		  inst_data->sess_file_name);
	  tj_log(inst_data, log_msg, c_red);
	  exit(EXIT_FAILURE);
	  }
	else
	  return -1;
	}

  /* load session file to memory */
  if (inst_data->sess_data)
	free(inst_data->sess_data);
  inst_data->sess_data = tj_malloc(inst_data->sess_data,
	  inst_data->sess_data_len + 1);
  fread(inst_data->sess_data, inst_data->sess_data_len, 1, sess_fp);
  fclose(sess_fp);
  return 0;
  }

void tj_log(struct tj_data *inst_data, char *message, int color)
  {
  if (inst_data->debug || inst_data->verbose)
	{
	if (inst_data->gui == 0)
	  printf("%s %s\n", "tcpjunk:", message);
	else
	  {
	  char gui_message[LINE_MAX];
	  sprintf(gui_message, "%s\n", message);
	  gui_add_log(gui_message, color);
	  }
	}
  }

void find_sections(struct tj_data *inst_data)
  {
  /* find all send buffers in session data */
  char *sects_buf = inst_data->sess_data;
  char *sects_buf_loc = sects_buf;
  inst_data->sect_idx = 0;
  char *start_sect_tag = "<sect>";
  char *end_sect_tag = "</sect>";

  /* if there are sections, enumerate them */
  while ((sects_buf_loc = memmem(sects_buf_loc, (inst_data->sess_data
	  + inst_data->sess_data_len - sects_buf_loc), start_sect_tag, strlen(
	  start_sect_tag))))
	{

	/* start tag found */
	sects_buf_loc += strlen(start_sect_tag);
	inst_data->sects[inst_data->sect_idx] = sects_buf_loc;

	/* get section end */
	if (!(sects_buf_loc = memmem(sects_buf_loc, sects_buf
		+ inst_data->sess_data_len - sects_buf_loc, end_sect_tag, strlen(
		end_sect_tag))))
	  {
	  /* end tag not found, using all buffer */
	  goto single_sect;
	  }

	/* end tag found */
	inst_data->sects_len[inst_data->sect_idx] = sects_buf_loc
		- inst_data->sects[inst_data->sect_idx];

	inst_data->sect_idx++;
	sects_buf_loc++;
	sects_buf = sects_buf_loc;

	}

  /* just a single section */
  single_sect:

  if (inst_data->sect_idx == 0)
	{
	inst_data->sects[inst_data->sect_idx] = sects_buf;
	inst_data->sects_len[inst_data->sect_idx] = inst_data->sess_data_len;
	inst_data->sect_idx++;
	}
  return;
  }

void set_auto_session_file(struct tj_data *inst_data)
  {
  /* choose a session file randomly */
  if (inst_data->auto_file == f_rand)
	{
	strcpy(inst_data->sess_file_name, inst_data->files_list[rand()
		% inst_data->auto_files_num]);
	}
  else
  /* choose a session file sequentially */
	{
	static int curr_file = 0;
	strcpy(inst_data->sess_file_name, inst_data->files_list[curr_file]);

	curr_file++;
	if (curr_file == inst_data->auto_files_num)
	  curr_file = 0;
	}

  if (inst_data->verbose)
	{
	sprintf(log_msg, "Using session file: %s", inst_data->sess_file_name);
	tj_log(inst_data, log_msg, c_black);
	}
  }

void load_file_list(struct tj_data *inst_data, const char *auto_file_dir)
  {
  int file_index = 0;
  struct dirent* dfile;
  char file_name[FILENAME_MAX];

  /* open directory */
  DIR *cur_dir = opendir(auto_file_dir);
  if (!cur_dir)
	{
	sprintf(log_msg, "error opening directory %s: %s", auto_file_dir, strerror(
		errno));
	tj_log(inst_data, log_msg, c_red);
	exit(EXIT_FAILURE);
	}

  /* get number of files */
  while ((dfile = readdir(cur_dir)) != NULL)
	{
	if (dfile->d_type != DT_DIR)
	  file_index++;
	}
  inst_data->files_list = (char **) malloc(file_index
	  * sizeof(*inst_data->files_list));
  if (inst_data->files_list == NULL)
	{
	printf("tcpjunk: memory allocation failed, terminating...\n");
	exit(EXIT_FAILURE);
	}

  /* store file names */
  int i = 0;
  rewinddir(cur_dir);
  while (((dfile = readdir(cur_dir)) != NULL))
	{
	if (dfile->d_type != DT_DIR)
	  {
	  sprintf(file_name, "%s/%s", auto_file_dir, dfile->d_name);
	  inst_data->files_list[i] = strdup(file_name);

	  if (inst_data->files_list[i] == NULL)
		{
		printf("tcpjunk: memory allocation failed, terminating...\n");
		exit(EXIT_FAILURE);
		}
	  i++;
	  }
	}

  /* close */
  if (closedir(cur_dir) == -1)
	{
	perror("tcpjunk: error closing directory ");
	exit(EXIT_FAILURE);
	}
  inst_data->auto_files_num = file_index;

  }

char *tj_malloc(char *buf, int buf_len)
  {
  buf = malloc(buf_len);
  if (buf == NULL)
	{
	printf("tcpjunk: memory allocation failed, terminating...\n");
	exit(EXIT_FAILURE);
	}
  else
	return buf;
  }

char *tj_realloc(char *buf, int buf_len)
  {
  buf = realloc(buf, buf_len);
  if (buf == NULL)
	{
	printf("tcpjunk: memory allocation failed, terminating...\n");
	exit(EXIT_FAILURE);
	}
  else
	return buf;
  }

void cleanup(int sig)
  {
  printf("\ntcpjunk: terminating...\n");
  exit(EXIT_SUCCESS);
  }

char *hex_dump(char *buf, int buf_len)
  {
  int buf_loc;
  int asc_loc, asc_offset;
  int line_loc = 0;
  static char *xbuf;
  if (xbuf)
	free(xbuf);
  xbuf = tj_malloc(xbuf, buf_len * 10 + 80);
  memset(xbuf, 0, buf_len * 10);
  int xbuf_loc = 0;

  for (buf_loc = 0; buf_loc < buf_len; buf_loc++)
	{
	line_loc++;

	/* start of line, print offset */
	if (line_loc == 1)
	  xbuf_loc += sprintf(xbuf + xbuf_loc, "%08x   ", buf_loc);

	/* print hex char */
	xbuf_loc += sprintf(xbuf + xbuf_loc, "%02x ", (unsigned char) buf[buf_loc]);

	/* print a space if in the middle */
	if (line_loc == 8)
	  xbuf_loc += sprintf(xbuf + xbuf_loc, " ");

	/* end of line or data */
	if (line_loc == 16 || buf_loc == buf_len - 1)
	  {
	  /* do some padding */
	  if (buf_loc == buf_len - 1 && line_loc != 16)
		{
		int temp_loc = line_loc;
		if (temp_loc < 8)
		  xbuf_loc += sprintf(xbuf + xbuf_loc, " ");

		while (temp_loc <= 15)
		  {
		  xbuf_loc += sprintf(xbuf + xbuf_loc, "   ");
		  temp_loc++;
		  }
		}

	  /* print a space between hex data and ASCII */
	  xbuf_loc += sprintf(xbuf + xbuf_loc, "  ");

	  /* print ASCII */
	  asc_offset = 0;
	  for (asc_loc = buf_loc - line_loc + 1; asc_loc <= buf_loc; asc_loc++)
		{
		if (isprint(buf[asc_loc]))
		  xbuf_loc += sprintf(xbuf + xbuf_loc, "%c", buf[asc_loc]);
		else
		  xbuf_loc += sprintf(xbuf + xbuf_loc, ".");
		asc_offset++;
		if (asc_offset == 8)
		  xbuf_loc += sprintf(xbuf + xbuf_loc, " ");
		}

	  /* new line */
	  xbuf_loc += sprintf(xbuf + xbuf_loc, "\n");
	  line_loc = 0;
	  }
	}

  return xbuf;
  }

void gen_new_src_ip(struct tj_data *inst_data, char *new_src_ip)
  {
  if (inst_data->ipv6)
	{
	char tmp_oct[16];
	char oct[8][16];
	int args;
	int oct_num;
	args = sscanf(new_src_ip,
		"%[^:]:%[^:]:%[^:]:%[^:]:%[^:]:%[^:]:%[^:]:%[^:]", oct[0], oct[1],
		oct[2], oct[3], oct[4], oct[5], oct[6], oct[7]);
	if (args < 8)
	  {
	  tj_log(inst_data, "Invalid source IPv6 address format", c_red);
	  if (inst_data->gui == 0)
		exit(EXIT_FAILURE);
	  }

	new_src_ip[0] = '\0';
	for (oct_num = 0; oct_num <= 7; oct_num++)
	  {

	  if (oct[oct_num][0] == 'x')
		{
		sprintf(tmp_oct, "%x", rand() % 65534);
		strcat(new_src_ip, tmp_oct);
		}
	  else
		strcat(new_src_ip, oct[oct_num]);

	  if (oct_num < 7)
		strcat(new_src_ip, ":");
	  }
	}
  else
	{
	char tmp_oct[8];
	char oct[4][8];
	int args;
	args = sscanf(new_src_ip, "%[^.].%[^.].%[^.].%[^.]", oct[0], oct[1],
		oct[2], oct[3]);
	if (args < 4)
	  {
	  tj_log(inst_data, "Invalid source IPv4 address format", c_red);
	  if (inst_data->gui == 0)
		exit(EXIT_FAILURE);
	  }

	new_src_ip[0] = '\0';
	for (args = 0; args <= 3; args++)
	  {
	  if (oct[args][0] == 'x')
		{
		sprintf(tmp_oct, "%d", rand() % 254);
		strcat(new_src_ip, tmp_oct);
		}
	  else
		strcat(new_src_ip, oct[args]);

	  if (args < 3)
		strcat(new_src_ip, ".");
	  }
	}
  }

void daemonize()
  {
  pid_t pid, sid;

  /* new child */
  pid = fork();
  if (pid < 0)
	{
	printf("tcpjunk: failed to daemonize\r\n");
	exit(EXIT_FAILURE);
	}

  if (pid > 0)
	{
	exit(EXIT_SUCCESS);
	}

  /* new session ID */
  sid = setsid();
  if (sid < 0)
	{
	printf("tcpjunk: failed to daemonize\r\n");
	exit(EXIT_FAILURE);
	}

  /* output redirection */
  freopen("/dev/null", "r", stdin);
  freopen("/dev/null", "w", stdout);
  freopen("/dev/null", "w", stderr);
  }
